/*
Connect PE0,PE1,PD5,PD4 together
ACO -> T0,T1,T4

T0    count the number of pulse -> odd/even
T1    disabled (inutile)
T4    generate the trigger output

PINS
----
  from PD7 to 3.3V with 15k resistor
  from PD7 to Gnd with 1k resistor

PB0   OLED DC
PB1   unused
PB2   OLED CS
PB3   OLED MOSI
PB4   MISO (unused)
PB5   OLED SCLK
PB6   unused
PB7   unused

PC0   DN  10
PC1   DN  1
PC2   UP  1
PC3   UP  10
PC4   ODD/~EVEN
PC5   unused

PD0   RXD0  serial port
PD1   TXD0
PD2   trig pulse output (hard)
PD3   odd/even
PD4   --- T0 in
PD5   --- T1 in
PD6   Video input
PD7   trig level at 0.2V

PE0   ACO
PE1   T4 input
PE2   vertical sync pulse
PE3   OLED reset

Principle of operation
1) The analog comparator separates the horizontal sync signal from the video signal
2) The ACO ISR gets the vertical sync signal at line 4 in the vertical sync pulse interval.
3) The ACO output is fed to T4
4) T0 decides which field: odd/even by counting the nb of µs between syncs and pipeline this until the vertical sync
5) T4 is used to generate the trigger output by counting the lines over 2 fields
6) The GUI selects which line by converting the line number and field into the sync count of T4
7) the comparator match output is the final trigger output fed to the oscilloscope
*/

#include <avr/io.h>

#include <SPI.h>  // Include SPI if you're using SPI
#include <SFE_MicroOLED.h>  // Include the SFE_MicroOLED library

//////////////////////////
// MicroOLED Definition //
//////////////////////////

#define PIN_RESET A7  // Connect RST to pin 9
#define PIN_DC    8  // Connect DC to pin 8
#define PIN_CS    10 // Connect CS to pin 10
#define DC_JUMPER 0 // Set to either 0 (SPI, default) or 1 (I2C) based on jumper, matching the value of the DC Jumper

//////////////////////////////////
// MicroOLED Object Declaration //
//////////////////////////////////
// Declare a MicroOLED object. The parameters include:
// 1 - Reset pin: Any digital pin
// 2 - D/C pin: Any digital pin (SPI mode only)
// 3 - CS pin: Any digital pin (SPI mode only, 10 recommended)

#define K_ODD_EVEN    0x0F
#define K_DN_10       0x1E
#define K_DN_1        0x1D
#define K_UP_1        0x1B
#define K_UP_10       0x17

#define CTR_BEGIN  1
#define CTR_END    262
#define TIMER4_TOP  (525+17)

MicroOLED oled(PIN_RESET, PIN_DC, PIN_CS);

volatile byte sync_counter;
volatile byte pipeline;
boolean odd;
byte t4_mode = 0x0E;

ISR(BADISR_vect)
{
}

void setup()
{
  CLKPR = 0x80;
  CLKPR = 0x00;

  sync_counter = 0;

  DDRB = 0xFF;
  DDRD = 0x0F;
  DDRE = 0b00001101;
  //PORTC as input
  DDRC = 0x20;

//--------------
// Analog Comparator setup
  // disable AIN0,1 digital input buffer
  DIDR1 = 0x03;

  // disable Analog Comparator Multiplexer Enable
  ADCSRB = 0;

  //Bit 0: ACO enable
  //Bits 0-7: unused
  ACSRB = 1;

  ACSR = bit(ACI)                 // clear Analog Comparator Interrupt Flag
         | bit(ACIE)                // set Analog Comparator Interrupt Enable
         | bit(ACIS1);              // select falling edge: ACIS1/ACIS0 Analog Comparator Interrupt Mode Select

//--------------
// Timer0 setup
  // for separating odd/even fields, we check the length of the last line, 
  // pipeline the decision until the time the vertical sync is generated.
  // The decision is made by timer 0, which gives 1 for a full line, 0 for a half line.
  // The pipeline is implemented by shifting a 8-bit register for 6,7,or 8 horizontal syncs.
  // H = 63.5µs
  // Assume power is enabled
  // Assume clkIO = 8MHz
  // TCNT0 value is read and pipelined
  // TCNT0 is cleared at each horizontal sync interrupt,
  // It counts clkIO
  
  TCCR0A = 0b00000000;  // normal op; OC0A,OC0B disconnected TOP=FF
  TCCR0B = 0b00000010;  // clkIO/8
  TIMSK0 = 0;           // no interrupts
  TCNT0 = 0;

  cli();

//--------------
// Timer4 setup
// fast mode 7
// scope trigger output
// COM4A = 3 rising
// COM4A = 2 falling
// TOP = 525 + 17
// double buffer -> written when timer clock cycle the TCNT4 matches TOP

  // The comparator register will then cover 2 fields and count from 1 to TOP
  TCCR4A = (t4_mode & 0x03) | 0b11000000;  //USE OC4A output (PD1)
  TCCR4B = ((t4_mode & 0x0C)<<1) | 0b00000110;	// ext clock on falling edge
  OCR4A = 535;
  ICR4 = TIMER4_TOP;
  TCNT4 = 0;

  //TIMSK4 = 0b00000011;

//--------------
// OLED setup
  oled.begin();    // Initialize the OLED

  oled.clear(ALL); // Clear the display's internal memory

  oled.setFontType(0);
  oled.setCursor(0, 40); // Set cursor to top-left
  oled.print("Trig v1.0");
  oled.display();  // Display what's in the buffer (splashscreen)

  displayCurrentField(1);
  sei();
}

void displayCurrentField(uint16_t myValue)
{
int16_t temp;

  oled.setFontType(3);

  /* Display comparator value */
  oled.clear(PAGE);     // Clear the page
  oled.setCursor(10, 0); // Set cursor to top-left
  oled.print(myValue);

  /* Display odd/even field */
  oled.setFontType(1);  // 8x16
  oled.setCursor(10, 32); // Set cursor to top-left
  if (odd) oled.print("odd");
  else oled.print("even");

  oled.display();       // Refresh the display
}

uint8_t getKeys()
{
  uint8_t t1, t2, t3;

  do
  {
  /* Wait for key release */
    t1 = PINC & 0x1f;
    t2 = PINC & 0x1f;
    t3 = PINC & 0x1f;
  } while ((t1 != 0x1f) || (t2 != 0x1f) || (t3 != 0x1f));

  do
  {
    /* Wait for key pressed */
    while ((PINC & 0x1f) == 0x1f);

    t1 = PINC & 0x1f;
    t2 = PINC & 0x1f;
    t3 = PINC & 0x1f;

    if ((t2 == t1) && (t3 == t2)) return t2;
  } while (true);
}

void loop()
{
  uint8_t keys;
//  const uint16_t table_value[9] = {535,536,538,540,1,3,5,7,9};
  const uint16_t table_value[9] = {536,538,540,542,1,3,5,7,9};
  uint16_t comparator;
  static uint16_t l_value = 1;

  keys = getKeys();

  switch (keys)
  {
    case K_ODD_EVEN:
      odd = !odd;
      break;
    case K_UP_1:
      l_value++;
      if (l_value > CTR_END) l_value = CTR_BEGIN;
      break;
    case K_UP_10:
      l_value += 10;
      if (l_value > CTR_END) l_value = CTR_BEGIN;
      break;
    case K_DN_1:
      l_value--;
      if (l_value < CTR_BEGIN) l_value = CTR_END;
      break;
    case K_DN_10:
      l_value -= 10;
      if (l_value < CTR_BEGIN) l_value = CTR_END;
      //No action on combined keys
  }
 
  displayCurrentField(l_value);

  if (odd)
  {
    if (l_value < 10) comparator = 264 + 2 * l_value;
    else comparator = 273 + l_value;
  }
  else /* even */
  {
    if (l_value > 9) comparator = 1 + l_value;
    else comparator = table_value[l_value - 1];
  }

  cli(); 
  OCR4A = comparator;

  // set TCNT4 to 3FE so that the OCR4A buffer is written to the OCR4A register
  TCNT4 = TIMER4_TOP;
  sei();
}
